#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingSchema.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "MappingMaterialization.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/MessageStream.h"

#include <sstream>
#include <memory>

pool::ObjectRelationalMappingPersistency::ObjectRelationalMappingPersistency( coral::ISchema& schema ):
  m_schema( schema ),
  m_storedMappings(),
  m_versionsLoaded( false ),
  m_tempMappingTable()
{}


pool::ObjectRelationalMappingPersistency::~ObjectRelationalMappingPersistency()
{
  for ( std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMapping = m_storedMappings.begin();
        iMapping != m_storedMappings.end(); ++iMapping ) {
    if ( iMapping->second ) delete iMapping->second;
  }
}

void
pool::ObjectRelationalMappingPersistency::initialize(bool forWriting){
  if(forWriting) createIfNecessaryTheMappingTables();
  const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();  
}

bool 
pool::ObjectRelationalMappingPersistency::removeMaterializedTables(const std::string& version, bool dryRun ){

  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  try {
    const pool::ObjectRelationalMapping& fullMapping = retrieveMapping( version );
    std::set<std::string> tableList;
    for(ObjectRelationalMapping::const_iterator iChild=fullMapping.begin();iChild!=fullMapping.end();iChild++)
    {
      makeTableList(iChild->second,tableList);
    }
    if(tableList.size()>0){
      log << coral::Info << " Dropping tables referenced by mapping version \"" << version << "\"."<<coral::MessageStream::endmsg;
    } else {
      log << coral::Info << " No table found referenced by mapping version \"" << version << "\"."<<coral::MessageStream::endmsg;
      return false;
    }
    
    for(std::set<std::string>::reverse_iterator iTable=tableList.rbegin();iTable!=tableList.rend();iTable++){

      // check if some other mapping version is referencing the same table
      std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() ).newQuery() );
      query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() );
      std::ostringstream condition;
      condition << pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() << " =:" <<
        pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable();
      coral::AttributeList condData;
      condData.extend<std::string>(pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable());
      coral::AttributeList::iterator iAttribute = condData.begin();
      iAttribute->data< std::string >() = *iTable;
      query->setCondition( condition.str(), condData );
      std::set<std::string> versions;
      coral::ICursor& cursor = query->execute();
      while ( cursor.next() ) {
        const coral::AttributeList& currentRow = cursor.currentRow();
        versions.insert(currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()].data<std::string>());
      }
      versions.erase( version );
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      if(m_schema.existsTable(*iTable)){
        if(versions.empty()){
          log << coral::Warning << "Dropping table \"" << *iTable <<"\"" << coral::MessageStream::endmsg;
          if(!dryRun) m_schema.dropTable(*iTable);
        } else{
          log << coral::Info << "Table \"" << *iTable <<"\" is currenlty used by other mapping versions." << coral::MessageStream::endmsg;
        }
      }
    }
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::removeMaterializedTables" );
  }
  return true;
}

bool
pool::ObjectRelationalMappingPersistency::removeMapping( const std::string& version )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  if ( ! m_versionsLoaded) loadMappingVersions();
  if ( m_storedMappings.find( version ) == m_storedMappings.end() ) {
    // Print an error here
    log << coral::Warning << "A mapping with version name \"" << version
        << "\" does not exist in the database. Cannot remove it." << coral::MessageStream::endmsg;
    return false;
  }
  
  // Remove the transient representation of the mapping
  std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMapping = m_storedMappings.find( version );
  if ( iMapping->second ) delete iMapping->second;
  m_storedMappings.erase( iMapping );

  // Remove all rows in the tables with the version.
  coral::AttributeList whereData;
  whereData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
  whereData.begin()->data<std::string>() = version;

  try {
    std::string conditionForClassVersionTable = pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() + " = :" + pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).dataEditor().deleteRows( conditionForClassVersionTable, whereData );
    std::string conditionForVersionTable = pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() + " = :" + pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).dataEditor().deleteRows( conditionForVersionTable, whereData );
    std::string conditionForElementTable = pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() + "= :" + pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() ;
    m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() ).dataEditor().deleteRows( conditionForElementTable, whereData );
  } catch ( const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::removeMapping" );    
  }
  return true;
}


bool
pool::ObjectRelationalMappingPersistency::storeMapping( const pool::ObjectRelationalMapping& mapping,
                                                        const std::string& classVersion,
                                                        std::string containerName  )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );

  // Check whether the mapping version already exists.
  if ( ! m_versionsLoaded) loadMappingVersions();
  if ( m_storedMappings.find( mapping.version() ) != m_storedMappings.end() ) {
    log << coral::Error << "A mapping with version name \"" << mapping.version()
        << "\" is already stored in the database. No new mapping is stored." << coral::MessageStream::endmsg;
    return false;
  }

  try {
    
    // Store an entry into the version table
    coral::ITable& versionTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() );
    coral::AttributeList dataForVersionTable;
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttribute = dataForVersionTable.begin();
    iAttribute->data< std::string >() = mapping.version();
    ++iAttribute;
    iAttribute->data< std::string >() = containerName;
    versionTable.dataEditor().insertRow( dataForVersionTable );

    // Store an entry into the class version table
    coral::ITable& classVersionTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() );
    coral::AttributeList dataForClassVersionTable;
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttr = dataForClassVersionTable.begin();
    iAttr->data< std::string >() = mapping.version();
    ++iAttr;
    iAttr->data< std::string >() = classVersion;
    ++iAttr;
    iAttr->data< std::string >() = containerName;

    std::string existingVersion = this->selectMappingVersion( classVersion, containerName );
    if(existingVersion.empty()){
      classVersionTable.dataEditor().insertRow( dataForClassVersionTable );
    } else {
      std::string setClause = pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable();
      std::string whereClause = pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" AND "+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
      classVersionTable.dataEditor().updateRows(setClause,whereClause,dataForClassVersionTable);
    }

    // Store the mapping elements.
    coral::ITable& mappingTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() );
    coral::AttributeList  dataBuffer;
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementIdFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() );
    dataBuffer.extend< int >( pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable() );

    dataBuffer[ pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable() ].data<std::string>()= mapping.version();

    unsigned int iElementIndex = 0;
    for ( pool::ObjectRelationalMapping::const_iterator iElement = mapping.begin();
          iElement != mapping.end(); ++iElement ) {
      std::stringstream elementId;
      elementId << iElementIndex;
      pool::ObjectRelationalMappingPersistency::storeMappingElement( elementId.str(),
                                                                     iElement->second,
                                                                     mappingTable.dataEditor(),
                                                                     dataBuffer );
      iElementIndex++;
    }
    
    m_storedMappings[ mapping.version() ] = 0;
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMapping" );    
  }
  return true;
}

std::map<std::string,std::string>
pool::ObjectRelationalMappingPersistency::retrieveMappingVersions() const
{
  std::map<std::string,std::string> mappingList;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string containerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
      std::string mappingVersion = currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>();
      mappingList.insert(std::make_pair(mappingVersion,containerName));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMappingVersions" );
  }
  return mappingList;
}

std::map<std::pair<std::string,std::string>,std::string>
pool::ObjectRelationalMappingPersistency::retrieveMappingForClassVersions() const{
  std::map<std::pair<std::string,std::string>,std::string> classVersionMap;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).newQuery() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string classVersion = currentRow[pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()].data<std::string>();
      std::string containerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable()].data<std::string>();
      std::string mappingVersion = currentRow[pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()].data<std::string>();
      classVersionMap.insert(std::make_pair(std::make_pair(classVersion,containerName),mappingVersion));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMappingForClassVersions" );
  }
  return classVersionMap;
}

std::string
pool::ObjectRelationalMappingPersistency::selectMappingVersion( const std::string& classVersion, const std::string& containerName ) const
{
  std::string result("");
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).newQuery() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() << " =:" <<
      pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() << " AND " <<
      pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
    condition << " =:" << pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();

    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = classVersion;
    ++iAttribute;
    iAttribute->data< std::string >() = containerName;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      result = currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>();
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::selectMappingVersion" );
  }
  return result;
}

void
pool::ObjectRelationalMappingPersistency::selectMapping(const std::string& version,
                                                        std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct) const
{
  coral::ITable& mappingTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() );
  try {
    std::auto_ptr<coral::IQuery> query(mappingTable.newQuery());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable()<<"= '"<<
      pool::ObjectRelationalMappingElement::primitiveMappingElementType()<<"' AND "<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList("T0."+pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOrderList("T0."+pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string scopeName = currentRow[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string variableName = currentRow[ pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string tableName = currentRow[ pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() ].data<std::string>();
      std::string columnName = currentRow[ pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable()].data<std::string>();
      mappingStruct.insert(std::make_pair(std::make_pair(scopeName,variableName),std::make_pair(tableName,columnName)));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::selectMapping" );
  }
}

std::string
pool::ObjectRelationalMappingPersistency::containerForMappingVersion( const std::string& mappingVersion) const {
  std::string currentContainerName("");
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = mappingVersion;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      currentContainerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerForMappingVersion" );
  }
  return currentContainerName;
}

std::vector<std::string>
pool::ObjectRelationalMappingPersistency::mappingsForContainer( const std::string& containerName ) const {
  std::vector<std::string> mappings;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = containerName;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      mappings.push_back(currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>());
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::mappingsForContainer" );
  }
  return mappings;
}

std::map<std::string, std::string>
pool::ObjectRelationalMappingPersistency::containerTableMap()
{
  std::map<std::string,std::string> containerTableMap;
  try {
    std::auto_ptr<coral::IQuery> query(m_schema.newQuery());
    query->addToTableList( pool::ObjectRelationalMappingSchema::mappingElementTable(),"T0");
    query->addToTableList( pool::ObjectRelationalMappingSchema::mappingVersionTable(), "T1");
    query->setDistinct();
    query->addToOutputList( "T0."+pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    query->addToOutputList( "T1."+pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable());
    std::ostringstream condition;
    condition << "T0."<<pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()<<"="<<
      "T1."<<pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    query->setCondition(condition.str(),condData);
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string tableName = currentRow[ "T0."+pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable()].data<std::string>();
      std::string containerName = currentRow[ "T1."+pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
      containerTableMap.insert(std::make_pair(tableName,containerName));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerTableMap" );
  }
  return containerTableMap;
}

const pool::ObjectRelationalMapping&
pool::ObjectRelationalMappingPersistency::retrieveMapping( const std::string& version ) const
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  pool::ObjectRelationalMapping* mappingPointer = 0;
  try {
    std::map< std::string, pool::ObjectRelationalMapping* >::const_iterator iMapping = m_storedMappings.find( version );
    if ( iMapping != m_storedMappings.end() ) {
      mappingPointer = iMapping->second;
    } else {
      m_storedMappings[version] = 0;
    }
    if ( ! mappingPointer ) {
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMapping( version );
      mappingPointer = m_storedMappings.find( version )->second;
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
  if(! mappingPointer ){
    throw pool::ObjectRelationalException( "Non-existing mapping version name",
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
  
  return *mappingPointer;
}

pool::ObjectRelationalMapping*
pool::ObjectRelationalMappingPersistency::exportMapping( const std::string& version ) const
{

  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  pool::ObjectRelationalMapping* mapping = 0;
  try {
    if ( ! m_versionsLoaded) {
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();
    }

    std::map< std::string, pool::ObjectRelationalMapping* >::const_iterator iMapping = m_storedMappings.find( version );
    if ( iMapping == m_storedMappings.end() ) {
      log << coral::Error << "No mapping with version name \"" << version << "\" is stored in the database." << coral::MessageStream::endmsg;
    } else {
      mapping = new pool::ObjectRelationalMapping( version );
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingInOneGo( version, mapping);
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
  if(!mapping){
    throw pool::ObjectRelationalException( "Non-existing mapping version name",
                                             "ObjectRelationalMappingPersistency::exportMapping" );
  }
  return mapping;
}

bool
pool::ObjectRelationalMappingPersistency::materializeMapping( const std::string& version,
                                                              bool evolve,
                                                              bool dryRun )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  // Checking whether the mapping exists.
  bool ret = false;
  try {
    if ( ! m_versionsLoaded) {
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();
    }
    std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMapping = m_storedMappings.find( version );
    if ( iMapping == m_storedMappings.end() ) {
      log << coral::Error << "Mapping version \"" << version << "\" does not exist. Not materializing the mapping..." << coral::MessageStream::endmsg;
      return false;
    }
    if ( ! iMapping->second ) {
      this->loadMapping( version );
      iMapping = m_storedMappings.find( version );
    }
    pool::ObjectRelationalAccess::MappingMaterialization materialization( m_schema );
    ret = materialization.materialize( *( iMapping->second ), evolve, dryRun );
    if(dryRun){
      const std::set<std::string>& logs = materialization.operationList();
      if(logs.empty()) log << coral::Warning << "# No changes on the schema are needed." << coral::MessageStream::endmsg;
      for( std::set<std::string>::const_iterator iLog = logs.begin();
           iLog!=logs.end();++iLog){
        log << coral::Warning << "# Schema change: "<<*iLog<<coral::MessageStream::endmsg;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::materializeMapping" );
  }
  return ret;
}

bool
pool::ObjectRelationalMappingPersistency::materializeMapping( const pool::ObjectRelationalMapping& mapping,
                                                              bool evolve,
                                                              bool dryRun )
{
  // Checking whether the mapping exists.
  bool ret = false;
  try {
    pool::ObjectRelationalAccess::MappingMaterialization materialization( m_schema );
    ret = materialization.materialize( mapping, evolve, dryRun );
    if(dryRun){
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      const std::set<std::string>& logs = materialization.operationList();
      if(logs.empty()) log << coral::Warning << "# No changes on the schema are required." <<coral::MessageStream::endmsg;
      for( std::set<std::string>::const_iterator iLog = logs.begin();
           iLog!=logs.end();++iLog){
        log << coral::Warning << "# Schema change: "<<*iLog<<coral::MessageStream::endmsg;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::materializeMapping" );
  }
  return ret;
}

bool
pool::ObjectRelationalMappingPersistency::assignMapping( const std::string& classVersion, const std::string& mappingVersion, const std::string& toContainerName)
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  // first check if the specified mapping is used by an other container
  std::string currentCont = containerForMappingVersion(mappingVersion);
  if(currentCont.empty()){
    log << coral::Error << "Mapping \""<<mappingVersion<<"\" for container \""<<toContainerName<<"\" does not exist in the database."<<coral::MessageStream::endmsg;
    return false;
  } else {
    if(currentCont != toContainerName){
      log << coral::Error << "Mapping \""<<mappingVersion<<"\" is already assigned to container \""<<currentCont<<"\"."<<coral::MessageStream::endmsg;
      return false;
    }
  }
  try {
    std::auto_ptr<coral::IQuery> query0( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
    query0->addToOutputList( pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable() );
    std::ostringstream cond;
    cond << pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()<< " =:"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttr = condData.begin();
    iAttr->data< std::string >() = mappingVersion;
    query0->setCondition( cond.str(), condData);
    std::string currentContainerName("");
    coral::ICursor& cursor0 = query0->execute();
    while ( cursor0.next() ) {
      const coral::AttributeList& currentRow = cursor0.currentRow();      
      currentContainerName = currentRow[pool::ObjectRelationalMappingSchema::containerNameFieldInMappingVersionTable()].data<std::string>();
    }
    if(currentContainerName.empty()){
      log << coral::Error << "Mapping version \""+mappingVersion+"\" has not been found in the database." << coral::MessageStream::endmsg;
      return false;
    } else {
      if(toContainerName != currentContainerName){
        log << coral::Error << "Mapping version=\""<<mappingVersion<<"\" is already assigned to container \""<<currentContainerName<<"\"." << coral::MessageStream::endmsg;
        return false;
      }
    }
    
    // first check if the classVersion and the container are already represented
    std::auto_ptr<coral::IQuery> query1( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() ).newQuery() );
    query1->addToOutputList( pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable();
    condition << " AND ";
    condition << pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
    coral::AttributeList conditionData;
    conditionData.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    conditionData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = conditionData.begin();
    iAttribute->data< std::string >() = classVersion;
    ++iAttribute;
    iAttribute->data< std::string >() = toContainerName;    
    query1->setCondition( condition.str(), conditionData );
    std::set<std::string> classVersions;
    coral::ICursor& cursor1 = query1->execute();
    std::string currentVersion("");
    while ( cursor1.next() ) {
      const coral::AttributeList& currentRow = cursor1.currentRow();
      currentVersion = currentRow[pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()].data<std::string>();
    }
    bool update = false;
    if(!currentVersion.empty()){
      if(currentVersion!=mappingVersion){
        update = true;
      } else {
        log << coral::Debug << "Mapping version=\""<<mappingVersion<<"\" is already assigned to container \""<<toContainerName<<"\" for class version \""<<
          classVersion<<"\"." << coral::MessageStream::endmsg;
        return true;
      }
    }
    
    coral::ITable& classVersionTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::classVersionTable() );
    coral::AttributeList inputData;
    inputData.extend<std::string>(pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable());
    inputData.extend<std::string>(pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable());
    inputData.extend<std::string>(pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iInAttr = inputData.begin();
    iInAttr->data< std::string >() = mappingVersion;
    ++iInAttr;
    iInAttr->data< std::string >() = classVersion;
    ++iInAttr;
    iInAttr->data< std::string >() = toContainerName;
    if(update){
      std::string setClause = pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::mappingVersionFieldInClassVersionTable();
      std::string whereClause = pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::classVersionFieldInClassVersionTable()+" AND "+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalMappingSchema::containerNameFieldInClassVersionTable();
      classVersionTable.dataEditor().updateRows( setClause,whereClause, inputData  );
    } else {
      classVersionTable.dataEditor().insertRow( inputData );
    }
    log << coral::Debug << "Mapping version \""<< mappingVersion << "\" for class version \""<<classVersion<<
      "\" has been assigned to pool container \"" << toContainerName << "\""<<coral::MessageStream::endmsg;
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::assignMapping" );
  }
  return true;
}

void pool::ObjectRelationalMappingPersistency::makeTableList(const ObjectRelationalMappingElement& element,std::set<std::string>& tableList)
{
  tableList.insert(element.tableName());
  for(ObjectRelationalMappingElement::const_iterator iChild=element.begin();iChild!=element.end();iChild++)
  {
    makeTableList(iChild->second,tableList);
  }
}

void
pool::ObjectRelationalMappingPersistency::createIfNecessaryTheMappingTables(){
  ObjectRelationalMappingSchema mappingSchema(m_schema);
  if(!mappingSchema.existTables()){
    mappingSchema.createTables();
  }
}

void
pool::ObjectRelationalMappingPersistency::loadMappingVersions()
{
  if ( m_versionsLoaded ) return;
  std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingVersionTable() ).newQuery() );
  query->addToOutputList( pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable() );
  coral::ICursor& cursor = query->execute();
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    m_storedMappings[ currentRow[pool::ObjectRelationalMappingSchema::versionFieldInMappingVersionTable()].data<std::string>()] = 0;
  }
  m_versionsLoaded = true;
}

void
pool::ObjectRelationalMappingPersistency::loadMappingInOneGo( const std::string& version, pool::ObjectRelationalMapping* mapping )
{
  std::vector< std::pair<std::string, MappingDescription> > topLevelObjects;
  coral::ITable& mappingTable = m_schema.tableHandle( pool::ObjectRelationalMappingSchema::mappingElementTable() );
  std::auto_ptr<coral::IQuery> query(mappingTable.newQuery());
  try {
    query->addToOutputList( pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable());
    query->addToOutputList( pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() );
    query->addToOutputList( pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalMappingSchema::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList( pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable());
    query->addToOrderList( pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable());
    query->addToOrderList( pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string elementType = currentRow[ pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() ].data<std::string>();
      std::string scopeName = currentRow[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string variableName = currentRow[ pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string variableType = currentRow[ pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() ].data<std::string>();
      std::string tableName = currentRow[ pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() ].data<std::string>();
      std::string columnName = currentRow[ pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable()].data<std::string>();
      MappingDescription mapDescr;
      mapDescr.variableType = variableType;
      mapDescr.elementType = elementType;
      mapDescr.tableName = tableName;
      mapDescr.columns = std::vector<std::string>(1,columnName);
      if(scopeName == " "){  
        topLevelObjects.push_back(std::make_pair(variableName, mapDescr) );
      } else {
        std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
        if(iMap==m_tempMappingTable.end()) {
          std::map<std::string,MappingDescription> entry;
          entry.insert(std::make_pair(variableName,mapDescr));
          m_tempMappingTable.insert(std::make_pair(scopeName,entry));
        } else {                    
          std::map<std::string,MappingDescription>::iterator iInnerMap = iMap->second.find(variableName);
          if(iInnerMap==iMap->second.end()){
            iMap->second.insert(std::make_pair(variableName,mapDescr));
          } else {
            iInnerMap->second.columns.push_back(columnName);
          }
        }
      
      }
    
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::loadMappingInOneGo" );
  }
  
  for(std::vector<std::pair<std::string, MappingDescription> >::const_iterator iObj=topLevelObjects.begin();iObj!=topLevelObjects.end();iObj++){
    ObjectRelationalMappingElement& element = mapping->appendElement(iObj->first,iObj->second.tableName);    
    element.setColumnNames(iObj->second.columns);
    recursiveStoreMappingElement(element,iObj->first);
  }
  m_tempMappingTable.clear();
}

void
pool::ObjectRelationalMappingPersistency::loadMapping( const std::string& version )
{
  
  pool::ObjectRelationalMapping* mapping = m_storedMappings[ version ] = new pool::ObjectRelationalMapping( version );

  loadMappingInOneGo( version, mapping);


}
 
void
pool::ObjectRelationalMappingPersistency::storeMappingElement( const std::string& topElementId,
                                                               const pool::ObjectRelationalMappingElement& element,
                                                               coral::ITableDataEditor& dataEditor,
                                                               coral::AttributeList& dataBuffer )
{
  int iParamIndex = 0;
  size_t columns = element.columnNames().size();
  for ( std::vector< std::string >::const_iterator iColumn = element.columnNames().begin();
        iColumn != element.columnNames().end(); ++iColumn, ++iParamIndex ) {
    std::stringstream rowId;
    rowId << topElementId;
    if( columns > 1 ) rowId << "[" << iParamIndex << "]";
    dataBuffer[ pool::ObjectRelationalMappingSchema::elementIdFieldInMappingElementTable() ].data<std::string>()= rowId.str();
    dataBuffer[ pool::ObjectRelationalMappingSchema::elementTypeFieldInMappingElementTable() ].data<std::string>()= pool::ObjectRelationalMappingElement::elementTypeAsString( element.elementType() );
    if ( element.scopeName().empty() ) {
      dataBuffer[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() ].data<std::string>()= " " ;
    } else {
      dataBuffer[ pool::ObjectRelationalMappingSchema::scopeNameFieldInMappingElementTable() ].data<std::string>()= element.scopeName();
    }
    dataBuffer[ pool::ObjectRelationalMappingSchema::variableNameFieldInMappingElementTable() ].data<std::string>()= element.variableName();
    dataBuffer[ pool::ObjectRelationalMappingSchema::variableParIndexFieldInMappingElementTable() ].data<int>() = iParamIndex;
    dataBuffer[ pool::ObjectRelationalMappingSchema::variableTypeFieldInMappingElementTable() ].data<std::string>()= element.variableType();
    dataBuffer[ pool::ObjectRelationalMappingSchema::tableNameFieldInMappingElementTable() ].data<std::string>() = element.tableName();
    dataBuffer[ pool::ObjectRelationalMappingSchema::columnNameFieldInMappingElementTable() ].data<std::string>() = *iColumn;
    try 
    {
      dataEditor.insertRow( dataBuffer );
    }
    catch (const coral::SchemaException& e)
    {
      throw pool::ObjectRelationalException( e.what(),
                                             "ObjectRelationalMappingPersistency::storeMappingElement" );
    }
  }

  unsigned int elementIndex = 0;
  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    std::stringstream elementId;
    elementId << topElementId << "_" << elementIndex;
    pool::ObjectRelationalMappingPersistency::storeMappingElement( elementId.str(),iElement->second, dataEditor, dataBuffer );
    elementIndex++;
  }
}

void
pool::ObjectRelationalMappingPersistency::recursiveStoreMappingElement( ObjectRelationalMappingElement& parentElement, const std::string& scopeName ){
  std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
  if(iMap!=m_tempMappingTable.end()){
    for(std::map<std::string,MappingDescription>::const_iterator iMapEntry=iMap->second.begin();iMapEntry!=iMap->second.end();iMapEntry++){
      std::string variableName = iMapEntry->first;
      ObjectRelationalMappingElement& element = parentElement.appendSubElement( iMapEntry->second.elementType,
                                                                                variableName,
                                                                                iMapEntry->second.variableType,
                                                                                iMapEntry->second.tableName );
      element.setColumnNames(iMapEntry->second.columns);
      std::string nextScope(scopeName);
      nextScope.append("::").append(variableName);
      recursiveStoreMappingElement( element, nextScope);
    }
    m_tempMappingTable.erase(scopeName);
  }
}

